from typing import Any, Dict, List

import math
import random

from gurun.node import Node

try:
    import cv2
    import numpy as np
except ImportError:
    raise ImportError(
        "cv2 is not installed. Please install it with `pip install opencv-python`."
    )

BORDER_PROPORTION = 0.1


class RectToPoint(Node):
    def __call__(self, rect: np.ndarray, *args: Any, **kwargs: Any) -> Any:
        if len(np.shape(rect)) == 1:
            return {"x": rect[0] + (rect[2] / 2), "y": rect[1] + (rect[3] / 2)}
        elif len(np.shape(rect)) == 2:
            return [[r[0] + (r[2] / 2), r[1] + (r[3] / 2)] for r in rect]
        else:
            raise ValueError("RectToPoint: Input must be a 1D or 2D array")


class NaturalRectToPoint(Node):
    def __transform(self, value: int, limit: int) -> int:
        value_start = value + (limit * BORDER_PROPORTION)
        value_end = (value + limit) - (limit * BORDER_PROPORTION)
        return math.floor(random.uniform(value_start, value_end))

    def __call__(self, rect: np.ndarray, *args: Any, **kwargs: Any) -> Any:
        if len(np.shape(rect)) == 1:
            return {
                "x": self.__transform(rect[0], rect[2]),
                "y": self.__transform(rect[1], rect[3]),
            }
        elif len(np.shape(rect)) == 2:
            return [
                [self.__transform(r[0], r[2]), self.__transform(r[1], r[3])]
                for r in rect
            ]
        else:
            raise ValueError("RectToPoint: Input must be a 1D or 2D array")



from typing import Any, List

import random

from gurun.node import Node, WrapperNode

try:
    import pyautogui
except ImportError:
    raise ImportError(
        "pyautogui is not installed. Please install it with `pip install pyautogui`."
    )

pyautogui.FAILSAFE = False

MOUSE_RANDOM_MOVES = [
    pyautogui.easeInOutQuad,
    pyautogui.easeInOutCubic,
    pyautogui.easeInOutQuart,
    pyautogui.easeInOutQuint,
    pyautogui.easeInOutSine,
    pyautogui.easeInOutExpo,
    pyautogui.easeInOutCirc,
    pyautogui.easeInElastic,
    pyautogui.easeOutElastic,
    pyautogui.easeInOutElastic,
    pyautogui.easeInBounce,
    pyautogui.easeOutBounce,
    pyautogui.easeInOutBounce,
]
MINIMUM_DURATION = 0.1
MAXIMUM_DURATION = 1.0


class Typewrite(WrapperNode):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(pyautogui.typewrite, *args, **kwargs)


class Scroll(Node):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(pyautogui.scroll, *args, **kwargs)


class Click(WrapperNode):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(pyautogui.click, *args, **kwargs)


class HotKey(WrapperNode):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(pyautogui.hotkey, *args, **kwargs)


class MoveTo(WrapperNode):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(pyautogui.moveTo, *args, **kwargs)


class DragRel(WrapperNode):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(pyautogui.dragRel, *args, **kwargs)


class MoveRel(WrapperNode):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(pyautogui.moveRel, *args, **kwargs)


class MultipleClicks(Click):
    def __call__(self, positions: List[List[int]], *args: Any, **kwargs: Any):
        for x, y in positions:
            super().__call__(x, y * args, **kwargs, **self._memory)


class NaturalClick(Click):
    def __call__(self, *args: Any, **kwargs: Any):
        return super().__call__(
            *args,
            tween=random.choice(MOUSE_RANDOM_MOVES),
            duration=random.uniform(MINIMUM_DURATION, MAXIMUM_DURATION),
            **kwargs,
            **self._memory,
        )


class MultipleNaturalClicks(NaturalClick):
    def __call__(self, positions: List[List[int]], *args: Any, **kwargs: Any):
        for x, y in positions:
            super().__call__(x, y * args, **kwargs, **self._memory)



from typing import Any

import time

from gurun.node import Node, WrapperNode


class Sleep(WrapperNode):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(time.sleep, *args, **kwargs)


class RaiseException(Node):
    def __init__(self, exception: Exception, *args: Any, **kwargs: Any):
        super().__init__(default_state=False, *args, **kwargs)
        self._exception = exception

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        raise self._exception


class Periodic(Node):
    def __init__(self, node: Node, interval: int, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._node = node
        self._interval = interval
        self._last_run = 0

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        if time.time() - self._last_run < self._interval:
            self._state = False
            return None

        self._output = self._node(*args, **kwargs, **self._memory)
        self._state = self._node.state
        if self.state:
            self._last_run = time.time()

        return self._output


class Wait(Node):
    def __init__(self, node: Node, timeout: int, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._node = node
        self._timeout = timeout

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        start = time.time()
        while time.time() - start < self._timeout:
            self._node(*args, **kwargs, **self._memory)
            if self._node.state:
                self._state = True
                self._output = self._node.output
                return self._node.output

        self._state = False
        return None
